package mhtree.benchmarking;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation;
import messif.algorithms.AlgorithmMethodException;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorNeuralNetworkL2;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.RankedAbstractObject;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import messif.operations.Approximate;
import messif.operations.query.ApproxKNNQueryOperation;
import messif.operations.query.KNNQueryOperation;
import mhtree.InsertType;
import mhtree.MHTree;
import mhtree.ObjectToNodeDistance;
import mhtree.SearchState;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.stream.Collectors;

public class RunBenchmark {
    private static final String dataset = "./data/middle_c.txt";

    public static void main(String[] args) throws IOException, NoSuchMethodException, AlgorithmMethodException, BucketStorageException, ClassNotFoundException {
        if (!(args.length == 0 || args.length == 5)) {
            throw new IllegalArgumentException("unexpected number of params");
        }

        AbstractRepresentation.PrecomputedDistances.COMPUTATION_THREADS = 16;

        List<LocalAbstractObject> objects;

        if (args.length != 5) {
            objects = loadDataset(RunBenchmark.dataset);

            DCtoRecall(
                    new MHTreeConfig(
                            10,
                            2,
                            InsertType.GREEDY,
                            ObjectToNodeDistance.NEAREST
                    ),
                    objects,
                    new int[]{1, 50, 100}
            );

            return;
        }

        objects = loadDataset(args[0]);
        int leafCapacity = Integer.parseInt(args[1]);
        int nodeDegree = Integer.parseInt(args[2]);
        InsertType insertType = args[3].equals("INCREMENTAL") ? InsertType.INCREMENTAL : InsertType.GREEDY;
        ObjectToNodeDistance objectToNodeDistance;

        switch (args[4]) {
            case "FURTHEST":
                objectToNodeDistance = ObjectToNodeDistance.FURTHEST;
                break;
            case "AVERAGE":
                objectToNodeDistance = ObjectToNodeDistance.AVERAGE;
                break;
            default:
                objectToNodeDistance = ObjectToNodeDistance.NEAREST;
                break;
        }

//        Collections.shuffle(objects);

        DCtoRecall(new MHTreeConfig(
                        leafCapacity,
                        nodeDegree,
                        insertType,
                        objectToNodeDistance
                ),
                objects,
                new int[]{1, 50}
        );
    }

    private static void DCtoRecall(MHTreeConfig config, List<LocalAbstractObject> objects, int[] ks) throws NoSuchMethodException, AlgorithmMethodException, BucketStorageException, RuntimeException {
        MHTree mhTree = new MHTree.Builder(objects, config.leafCapacity, config.nodeDegree)
                .insertType(config.insertType)
                .objectToNodeDistance(config.objectToNodeDistance)
                .build();

        mhTree.printStatistics();

        int maxK = Arrays.stream(ks)
                .max()
                .orElseThrow(() -> new RuntimeException("bad max k"));

        // Precomputed NNs for each object
        HashMap<String, List<RankedAbstractObject>> URIToKNNs = new HashMap<>(objects.size());

        for (LocalAbstractObject object : objects) {
            KNNQueryOperation knnQueryOperation = new KNNQueryOperation(object, maxK);
            mhTree.executeOperation(knnQueryOperation);

            List<RankedAbstractObject> rankedKNNObjects = new ArrayList<>(knnQueryOperation.getAnswerCount());
            for (RankedAbstractObject knnObj : knnQueryOperation) {
                rankedKNNObjects.add(knnObj);
            }

            URIToKNNs.put(object.getLocatorURI(), rankedKNNObjects);
        }

        System.out.println("leafCapacity,nodeDegree,objectToNodeDistance,k,distanceComputations,recall (min),recall (avg),recall (med),recall (max)");

        double minRecall = 0;
        long distanceComputations = 0;

        for (int k : ks) {
            List<ApproxKNNQueryOperation> operations = objects
                    .stream()
                    .map(object -> new ApproxKNNQueryOperation(
                            object,
                            k,
                            1000,
                            Approximate.LocalSearchType.ABS_DC_COUNT,
                            LocalAbstractObject.UNKNOWN_DISTANCE))
                    .collect(Collectors.toList());

            operations.forEach(operation -> operation.suppData = new SearchState(null, null, false));

            System.out.println(
                    String.join(",",
                            String.valueOf(config.leafCapacity),
                            String.valueOf(config.nodeDegree),
                            String.valueOf(config.objectToNodeDistance),
                            String.valueOf(k),
                            String.valueOf(distanceComputations),
                            String.format("%.2f,%.2f,%.2f,%.2f", 0f, 0f, 0f, 0f)
                    )
            );

            distanceComputations += 1000;

            List<Double> recalls = new ArrayList<>(operations.size());
            for (int i = 0; i < operations.size(); i++) {
                recalls.add(0d);
            }

            while (minRecall != 1) {
                for (int i = 0; i < operations.size(); i++) {
                    if (recalls.get(i) != 1d) {
                        ApproxKNNQueryOperation operation = operations.get(i);

                        mhTree.executeOperation(operation);

                        recalls.set(i, PerformanceMeasures.measureRecall(operation, URIToKNNs));
                    }
                }

                Stats recallStats = new Stats(new ArrayList<>(recalls));

                System.out.println(String.join(",",
                        String.valueOf(config.leafCapacity),
                        String.valueOf(config.nodeDegree),
                        String.valueOf(config.objectToNodeDistance),
                        String.valueOf(k),
                        String.valueOf(distanceComputations),
                        String.format("%.2f,%.2f,%.2f,%.2f",
                                recallStats.getMin(),
                                recallStats.getAverage(),
                                recallStats.getMedian(),
                                recallStats.getMax())));

                minRecall = recallStats.getMin();
                distanceComputations += 1000;
            }

            minRecall = 0;
            distanceComputations = 0;
        }
    }

    private static List<LocalAbstractObject> loadDataset(String path) throws IOException {
        return new AbstractObjectList<>(new StreamGenericAbstractObjectIterator<>(ObjectFloatVectorNeuralNetworkL2.class, path));
    }
}
